from distutils.version import LooseVersion
import fnmatch
import json
from urllib.request import urlopen

from .asset_manager import download_assets
from .log import logger
# from .filesystem import find_files


def upload(*globs, dry_run=False):
    # files = find_files(globs)
    # logger().info("Uploading files to Github: %s",
    #               ', '.join(("'{}'".format(file) for file in files)),
    # )
    #
    # if dry_run:
    #     logger().info('Skipping GitHub upload step since this is a dry run.')
    # else:
    #     pass

    raise NotImplementedError('github.upload() is not yet implemented')


def _relevant_asset(asset, file_pattern):
    file_patterns = file_pattern.split(',')
    for pattern in file_patterns:
        if fnmatch.fnmatch(asset['name'], pattern):
            return True
    return False


def _get_release_info(repo, name, draft=False, prerelease=False):
    if '/' not in repo:
        raise ValueError(
            "repo must be of format <user>/<repo>, got '{}'".format(repo),
        )

    log = logger()
    url = 'https://api.github.com/repos/{}/releases'.format(repo)
    req = urlopen(url).read().decode()
    data = json.loads(req)
    releases = []

    for release in data:
        if not draft and release['draft']:
            log.info("Discarding draft release '%s'", release['name'])
        elif not prerelease and release['prerelease']:
            log.info("Discarding prerelease '%s'", release['name'])
        else:
            releases.append(release)

    try:
        if name == 'latest':
            release = max(
                releases,
                key=lambda x: LooseVersion(x['tag_name'].lstrip('v')),
            )
            log.info("Selected release '%s' as latest", release['name'])
        else:
            release = list(filter(lambda x: x['tag_name'] == name, releases))[0]

    except (IndexError, ValueError) as e:
        raise RuntimeError("No such Github release: '{}'".format(name)) from e

    return release


def download(repo, release, file_pattern, directory):
    release_info = _get_release_info(repo, release, file_pattern)
    assets = filter(lambda x: _relevant_asset(x, file_pattern),
                    release_info['assets'])
    download_assets(assets, directory, url_key='browser_download_url')



from pathlib import Path
from signal import Signals
import subprocess

import toml

from . import builder
from . import github
from . import pypi
from .filesystem import try_delete
from .log import logger


DOWNLOAD_SOURCES = {
    'gh': github,
    'github': github,
    'pypi': pypi.PRODUCTION,
    'pypi-test': pypi.TESTING,
}


def build():
    builder.dist()
    builder.zipapp()


def clean():
    try_delete("./build")
    try_delete("./dist")
    for name in Path.cwd().glob('*.egg-info'):
        if name.is_dir():
            try_delete(name)


def download(package, release_tag, file_pattern, directory):
    if file_pattern is None or len(file_pattern) == 0:
        raise ValueError('file_pattern must be non-empty.')

    if ':' not in package:
        raise ValueError('Invalid package/repository -- no source given.')

    source, package = package.split(':')

    if source not in DOWNLOAD_SOURCES.keys():
        raise ValueError('Invalid package/repository -- unknown source given.')

    source = DOWNLOAD_SOURCES[source]
    source.download(package, release_tag, file_pattern, directory)


def release(test_pypi, dry_run):
    if test_pypi:
        pypi_instance = pypi.TESTING
    else:
        pypi_instance = pypi.PRODUCTION
    pypi_instance.upload('./dist/*.tar.gz', './dist/*.whl', dry_run=dry_run)

    # if 'github' in args:
    #     github.upload('./dist/*.pyz', dry_run=dry_run)


def run(alias):
    pyproject = toml.load('pyproject.toml')

    try:
        command = pyproject['tool']['bork']['aliases'][alias]
    except KeyError:
        raise RuntimeError("No such alias: '{}'".format(alias))

    logger().info("Running '%s'", command)

    try:
        subprocess.run(command, check=True, shell=True)

    except subprocess.CalledProcessError as error:
        if error.returncode < 0:
            signal = Signals(- error.returncode)
            msg = "command '{}' exited due to signal {} ({})".format(
                error.cmd, signal.name, signal.value,
            )

        else:
            msg = "bork: command '{}' exited with error code {}".format(
                error.cmd, error.returncode,
            )

        raise RuntimeError(msg) from error



import inspect
import logging
import sys

import click
from click import BadParameter

from . import build as _build
from . import clean as _clean
from . import download as _download
from . import release as _release
from . import run as _run
from . import DOWNLOAD_SOURCES  # noqa: I100
from .log import logger


DOWNLOAD_SOURCES_STR = ' '.join(DOWNLOAD_SOURCES)


@click.group()
def cli():
    pass


@cli.command()
def build():
    _build()


@cli.command()
def clean():
    _clean()


@cli.command()
@click.option('--files', default='*.pyz',
              help='Comma-separated list of filenames to download. Supports '
                   'wildcards (* = everything, ? = any single character).')
@click.option('--directory', default='downloads',
              help='Directory to save files in. Created if missing.')
@click.argument('package', nargs=1)
@click.argument('release_tag', nargs=1, default='latest')
def download(files, directory, package, release_tag):
    # NOTE: We change the order of the arguments here, to move away from
    #       what makes sense on a CLI interface to what makes sense in a
    #       Python interface.
    try:
        _download(package, release_tag, files, directory)
    except ValueError as error:
        raise BadParameter(error)


@cli.command()
@click.option('--test-pypi', is_flag=True, default=False,
              help='Release to test.pypi.org instead of pypi.org.')
@click.option('--dry-run', is_flag=True, default=False,
              help="Don't actually release, just show what a release would do.")
def release(test_pypi, dry_run):
    _release(test_pypi, dry_run)


@cli.command()
@click.argument('alias', nargs=1)
def run(alias):
    _run(alias)


def main():
    verbose = '--verbose' in sys.argv
    if verbose:
        sys.argv.remove('--verbose')

    try:
        import coloredlogs  # pylint: disable=import-outside-toplevel
        coloredlogs.install(
            level=logging.INFO if verbose else logging.WARNING,
            fmt="%(name)s %(levelname)s %(message)s",
        )

    except ModuleNotFoundError:
        logging.basicConfig(
            level=logging.INFO if verbose else logging.WARNING,
            format="%(name)s %(levelname)s %(message)s",
        )

    try:
        cli()
    except RuntimeError as err:
        thrower = inspect.trace()[-1]
        log = logger(thrower)

        (log.exception if verbose else log.error)(str(err))

        sys.exit(1)



from twine.cli import dispatch as twine_upload

from .asset_manager import download_assets
from .filesystem import find_files
from .log import logger
from .pypi_api import get_download_info


class PypiHandler:
    def __init__(self, download_url, upload_url):
        self.download_url = download_url
        self.upload_url = upload_url

    def upload(self, *globs, dry_run=False):
        files = find_files(globs)
        logger().info(
            "Uploading files to PyPI instance '%s': %s",
            self.upload_url,
            ', '.join(("'{}'".format(file) for file in files)),
        )

        if dry_run:
            logger().info('Skipping PyPI upload step since this is a dry run.')
        else:
            twine_upload(['upload', '--repository-url', self.upload_url, *files])

    def download(self, package, release, file_pattern, directory):
        asset_list = get_download_info(self.download_url, package, release,
                                       file_pattern)
        download_assets(asset_list, directory)


PRODUCTION = PypiHandler('https://pypi.org/simple/',
                         'https://upload.pypi.org/legacy/')
TESTING = PypiHandler('https://test.pypi.org/simple/',
                      'https://test.pypi.org/legacy/')
